r"""
This module provides some tests of the integrator to known integrals.

Note, these are not the transformations, just the plain integrals,  :math:`\int_0^\infty f(x) J_\nu(x) dx`

There is a corresponding notebook in devel/ that runs each of these functions through a grid of N and h,
showing the pattern of accuracy. This could be useful for finding the correct numbers to choose for these
for other unknown functions.
"""

import pytest

import numpy as np
from scipy.special import gamma, gammainc, gammaincc, k0

from hankel import HankelTransform


def gammainc_(a, x):
    return gamma(a) * gammainc(a, x)


def gammaincc_(a, x):
    return gamma(a) * gammaincc(a, x)


def test_nu0_f_unity():
    """
    Test f(x) = 1, nu=0

    This test is done in the Ogata (2005) paper, section 5"
    """

    ht = HankelTransform(nu=0, N=50, h=0.1)
    ans = ht.integrate(lambda x: 1, False, False)
    print("Numerical Result: ", ans, " (required %s)" % 1)
    assert np.isclose(ans, 1, rtol=1e-3)


def test_nu0_f_x_on_x2():
    """
    Test f(x) = x/(x**2 + 1), nu=0

    This test is done in the Ogata (2005) paper, section 5"
    """
    ht = HankelTransform(nu=0, N=50, h=10 ** -1.5)

    ans = ht.integrate(lambda x: x / (x ** 2 + 1), False, False)
    print("Numerical Result: ", ans, " (required %s)" % k0(1))
    assert np.isclose(ans, k0(1), rtol=1e-3)


def test_nu0_f_x2():
    """
    Test f(x) = x^2, nu=0

    Result on wikipedia
    """
    ht = HankelTransform(nu=0, N=100, h=10 ** -1.5)

    ans = ht.integrate(lambda x: x ** 2, False, False)
    print("Numerical Result: ", ans, " (required -1)")
    assert np.isclose(ans, -1, rtol=1e-3)


def test_nu0_x4():
    """
    Result on wikipedia
    """
    ht = HankelTransform(nu=0, N=150, h=10 ** -1.5)
    ans = ht.integrate(lambda x: x ** 4, False, False)
    print("Numerical Result: ", ans, " (required 9)")
    assert np.isclose(ans, 9, rtol=1e-3)


def test_nu0_1_on_sqrt_x():
    """
    Result on wikipedia
    """
    # NOTE: this is REALLY finnicky!! (check devel/)
    ht = HankelTransform(nu=0, N=160, h=10 ** -3.5)
    ans = ht.integrate(lambda x: 1.0 / np.sqrt(x), False, False)
    m = -1.5
    anl = 2 ** (m + 1) * gamma(m / 2 + 1) / gamma(-m / 2)

    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)


def test_nu0_x_on_sqrt_x2_pz2():
    """
    Result on wikipedia
    """
    # Note that the number required is highly dependent on z .... smaller z is harder.
    ht = HankelTransform(nu=0, N=50, h=10 ** -1.3)

    z = 1
    ans = ht.integrate(lambda x: x / np.sqrt(x ** 2 + z ** 2), False, False)
    anl = np.exp(-z)
    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)


def test_nu0_f_gauss():
    """
    Result on wikipedia
    """
    z = 2
    ht = HankelTransform(nu=0, N=50, h=0.01)

    ans = ht.integrate(lambda x: x * np.exp(-0.5 * z ** 2 * x ** 2), False, False)
    anl = 1.0 / z ** 2 * np.exp(-0.5 / z ** 2)
    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)


@pytest.mark.parametrize(
    "s, nu, N, h",
    [
        [0, 1, 50, 0.05],
        [0, 2, 50, 0.05],
        [0.5, 1, 50, 0.05],
        [-2, 2, 600, 10 ** -2.6],  # This is pretty finnicky
        [-0.783, 1, 50, 0.05],
    ],
)
def test_nu_varying_powerlaw(s, nu, N, h):
    ht = HankelTransform(nu=nu, N=N, h=h)

    ans = ht.integrate(lambda x: x ** (s + 1), False, False)
    anl = 2 ** (s + 1) * gamma(0.5 * (2 + nu + s)) / gamma(0.5 * (nu - s))

    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)


@pytest.mark.parametrize(
    "s, nu, N, h", [[0.5, 1, 50, 0.05], [0.783, 1, 50, 0.05], [1.0, 0.5, 500, 0.01]],
)
def test_nu_varying_gamma_mod(s, nu, N, h):
    ht = HankelTransform(nu=nu, N=N, h=h)

    ans = ht.integrate(
        lambda x: x ** (nu - 2 * s + 1) * gammainc_(s, x ** 2), False, False
    )
    anl = 0.5 ** (2 * s - nu - 1) * gammaincc_(1 - s + nu, 0.25)

    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)



"""
This module provides tests that ensure that the forward then backward transform ends up
with the original.

For simplicity and because of common use (for me at least), we do this with a broken
powerlaw, which turns over at some x0. This *can* ensure convergence of the integral, as
long as the power-law exponents are chosen correctly (some here are commented out
because they don't fit the convergence criteria at high x).

There is a corresponding notebook in devel/ that runs each of these functions through a
grid of N and h, showing the pattern of accuracy. This could be useful for finding the
correct numbers to choose for these for other unknown functions.
"""

import pytest

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as Spline

from hankel import SymmetricFourierTransform

r = np.array([0.1, 1, 10.0])


@pytest.mark.parametrize(
    "s, t, x0, ndim, N, h, r",
    [
        [0, 1, 1, 2, 1100, 10 ** -2.5, r],
        [0, 2, 1, 2, 800, 10 ** -2.6, r],
        [0, 3, 1, 2, 1500, 10 ** -3.5, r[:2]],  # Doesn't work for high k!
        [1, 2, 1, 2, 250, 10 ** -2, r],
        [1, 3, 1, 2, 1200, 0.0008, r],
        [2, 3, 1, 2, 1200, 10 ** -2.5, r],
        [0, 1, 10, 2, 1200, 10 ** -2.6, r[1:]],  # Doesn't work for low k
        [0, 2, 10, 2, 250, 10 ** -2.0, r[1:]],  # Doesn't work for low k
        [0, 3, 10, 2, 1000, 10 ** -3, r],
        [1, 2, 10, 2, 1100, 10 ** -2.5, r],
        [1, 3, 10, 2, 250, 10 ** -2.2, r[1:]],  # Doesn't work for low k
        [2, 3, 10, 2, 700, 10 ** -2.3, r[1:]],  # Doesn't work for low k
        [0, 2, 1, 3, 700, 10 ** -2.4, r],
        [0, 3, 1, 3, 250, 10 ** -2, r],
        [1, 3, 1, 3, 2000, 10 ** -2.9, r],
        [0, 2, 10, 3, 2200, 10 ** -2.8, r],
        [0, 3, 10, 3, 400, 10 ** -2.4, r],
        [1, 3, 10, 3, 5000, 10 ** -3, r],  # Bad at low k
    ],
)
def test_roundtrip_broken_pl(s, t, x0, ndim, N, h, r):
    def f(x):
        return x ** s / (x ** t + x0)

    ht = SymmetricFourierTransform(ndim=ndim, N=N, h=h)

    k = np.logspace(
        np.log10(ht.x.min() / r.max() / 10.0),
        np.log10(10 * ht.x.max() / r.min()),
        1000,
    )
    resk = ht.transform(f, k, False)
    spl = Spline(k, resk)
    res = ht.transform(spl, r, False, inverse=True)

    print("Analytic: ", f(r), ", Numerical: ", res)
    assert np.allclose(res, f(r), atol=1e-2)



# -*- coding: utf-8 -*-
r"""Tools for Hankel transformations."""

import numpy as np
from mpmath import fp as mpm
from scipy.special import gamma, j0, j1, jn
from scipy.special import jn_zeros as _jn_zeros
from scipy.special import jv, yv

SRPI2 = np.sqrt(np.pi / 2.0)


def psi(t):
    """Compute the variable transform from Ogata 2005."""
    return t * np.tanh(np.pi * np.sinh(t) / 2)


def d_psi(t):
    """Compute the derivative of the variable transform from Ogata 2005."""
    t = np.array(t, dtype=float)
    a = np.ones_like(t)
    mask = t < 6
    t = t[mask]
    a[mask] = (np.pi * t * np.cosh(t) + np.sinh(np.pi * np.sinh(t))) / (
        1.0 + np.cosh(np.pi * np.sinh(t))
    )
    return a


def weight(nu, zeros):
    """Get weights for the summation in the hankel transformation."""
    return yv(nu, np.pi * zeros) / kernel(np.pi * zeros, nu + 1)


def roots(N, nu):
    """Get the first N Roots of the Bessel J(nu) functions divided by pi."""
    if np.isclose(nu, np.floor(nu)):
        return _jn_zeros(nu, N) / np.pi
    if np.isclose(nu, 0.5):
        # J(0.5) is just sqrt(2/(x*pi))*sin(x)
        return np.arange(1, N + 1)
    if np.isclose(nu, -0.5):
        # J(-0.5) is just sqrt(2/(x*pi))*cos(x)
        return np.arange(1, N + 1) - 0.5
    return np.array([mpm.besseljzero(nu, i + 1) for i in range(N)]) / np.pi


def j_lim(nu):
    """
    Compute the timit factor of Bessel J(nu, 0) = 0.5 ** nu / Gamma(nu + 1).

    Parameters
    ----------
    nu : float
        Order of the Bessel function.

    Returns
    -------
    float
        The factor.
    """
    return 0.5 ** nu / gamma(nu + 1)


def kernel(x, nu, alt=False):
    """
    Compute kernel functions for the hankel transformation.

    J(nu, x) or for alt=True: J(nu, x) * sqrt(x).

    Parameters
    ----------
    x : array-like
        input values.
    nu : int or float
        order of the bessel function.
    alt : bool, optional
        Whether the alternative defintion of the hankel transform should be
        used: J(nu, x)*sqrt(x). The default is False.

    Returns
    -------
    array-like
        The needed function for the hankel transformation.

    Notes
    -----
    J(nu, x) is approximately (x/2)^nu / Gamma(nu+1) for small x.
    """
    if alt:
        if np.isclose(nu, 0):
            return j0(x) * np.sqrt(x)
        if np.isclose(nu, 1):
            return j1(x) * np.sqrt(x)
        if np.isclose(nu, 0.5):  # J[0.5] = sqrt(2/(x*pi))*sin(x)
            return np.sin(x) / SRPI2
        if np.isclose(nu, -0.5):  # J[-0.5] = sqrt(2/(x*pi))*cos(x)
            return np.cos(x) / SRPI2
        if np.isclose(nu, np.floor(nu)):
            return jn(int(nu), x) * np.sqrt(x)
        return jv(nu, x) * np.sqrt(x)
    if np.isclose(nu, 0):
        return j0(x)
    if np.isclose(nu, 1):
        return j1(x)
    if np.isclose(nu, np.floor(nu)):
        return jn(int(nu), x)
    return jv(nu, x)


def safe_power(x, p):
    """
    Safely calculate x**p.

    Parameters
    ----------
    x : array-like
        value.
    p : float
        exponent.

    Returns
    -------
    array-like
        The result x**p.
    """
    return np.ones_like(x) if np.isclose(p, 0) else np.array(x ** p)


def get_x(h, zeros):
    """Get the arguments for the integrand."""
    return np.pi * psi(h * zeros) / h


def fourier_norm(a, b, ndim, inverse=False):
    r"""Calculate fourier-pair normalisations."""
    if inverse:
        return np.sqrt(np.abs(b) / (2 * np.pi) ** (1 + a)) ** ndim
    return np.sqrt(np.abs(b) / (2 * np.pi) ** (1 - a)) ** ndim


def dim_to_nu(ndim):
    """Calculate nu for the Hankel Transformation."""
    # keep int-type in python 3
    if np.isclose(ndim % 2, 0):
        return int(ndim) // 2 - 1
    return ndim / 2.0 - 1


def get_h(
    f,
    nu,
    K=None,
    cls=None,
    hstart=0.05,
    hdecrement=2,
    atol=1e-3,
    rtol=1e-3,
    maxiter=15,
    inverse=False,
):
    """
    Determine the largest value of h which gives a converged solution.

    Parameters
    ----------
    f : callable
        The function to be integrated/transformed.
    nu : float
        Either the order of the transformation, or the number of dimensions
        (if `cls` is a :class:`SymmetricFourierTransform`)
    K : float or array-like, optional
        The scale(s) of the transformation.
        If None, assumes an integration over f(x)J_nu(x) is desired.
        It is recommended to use a down-sampled K
        for this routine for efficiency. Often a min/max is enough.
    cls : :class:`HankelTransform` subclass, optional
        Either :class:`HankelTransform` or a subclass,
        specifying the type of transformation to do on `f`.
    hstart : float, optional
        The starting value of h.
    hdecrement : float, optional
        How much to divide h by on each iteration.
    atol, rtol : float, optional
        The tolerance parameters, passed to `np.isclose`,
        defining the stopping condition.
    maxiter : int, optional
        Maximum number of iterations to perform.
    inverse : bool, optional
        Whether to treat as an inverse transformation.

    Returns
    -------
    h : float
        The h value at which the solution converges.
    res : scalar or tuple
        The value of the integral/transformation using the returned h --
        if a transformation, returns results at K.
    N : int
        The number of nodes necessary in the final calculation.
        While each iteration uses N=pi/h, the returned N checks
        whether nodes are numerically zero above some threshold.

    Notes
    -----
    This function is not completely general. The function `f` is assumed to be
    reasonably smooth and non-oscillatory.

    The idea is to use successively smaller values of *h*, with N=pi/h on each
    iteration, until the result betweeniterations becomes stable.
    """
    # prevent circular imports
    from hankel.hankel import HankelTransform

    cls = HankelTransform if cls is None else cls

    if hstart >= 1:
        raise ValueError("h should never be greater than unity")

    # First, ensure that *some* of the values are non-zero
    i = 0
    while (
        np.any(
            np.all(
                cls(nu, h=hstart, N=int(np.pi / hstart))._get_series(
                    f, 1 if K is None else K
                )
                == 0,
                axis=-1,
            )
        )
        and i < maxiter
    ):
        hstart /= hdecrement
        i += 1

    if i == maxiter:
        raise Exception("Maxiter reached while checking for non-zero values")

    if K is None:  # Do a normal integral of f(x)J_nu(x)
        K = 1

        def getres(h):
            """Get the result."""
            return cls(nu, h=h, N=int(np.pi / h)).transform(
                lambda x: f(x) / x, k=K, ret_err=False, inverse=inverse
            )

    else:  # Do a transform at k=K

        def getres(h):
            """Get the result."""
            return cls(nu, h=h, N=int(np.pi / h)).transform(
                f, k=K, ret_err=False, inverse=inverse
            )

    res = getres(hstart)
    res2 = 2 * res + 10

    while not np.allclose(res, res2, atol=atol, rtol=rtol) and i < maxiter:
        i += 1
        hstart /= hdecrement
        res2 = 1 * res
        res = getres(hstart)

    if i == maxiter:
        raise Exception("Maxiter reached while checking convergence")

    # Can do some more trimming of N potentially, by seeing where f(x)~0.
    def consecutive(data, stepsize=1):
        """Split into arrays of consecutive zeros."""
        return np.split(data, np.where(np.diff(data) != stepsize)[0] + 1)

    hstart *= hdecrement

    x = cls(nu, h=hstart, N=int(np.pi / hstart)).x
    lastk = np.where(f(x / np.max(K)) == 0)[0]
    if len(lastk) > 1:
        # if there are any that are zero,
        # and if there are more than 1 in a row
        # (otherwise might just be oscillatory)
        lastk = consecutive(lastk)  # split into arrays of consecutive zeros
        if len(lastk[-1]) == 1:
            lastk = int(np.pi / hstart)
        else:
            lastk = lastk[-1][0]
    else:  # otherwise set back to N
        lastk = int(np.pi / hstart)

    return hstart, res2, lastk



"""
This module provides some tests for the Symmetric Fourier Transforms.

Note that most of the tests in here are mirrored in test_known_integrals, in which k is set to 1.

There is a corresponding notebook in devel/ that runs each of these functions through a grid of N and h,
showing the pattern of accuracy. This could be useful for finding the correct numbers to choose for these
for other unknown functions.
"""

import pytest

import numpy as np
from scipy.special import gamma, gammainc, gammaincc

from hankel import SymmetricFourierTransform


def gammainc_(a, x):
    return gamma(a) * gammainc(a, x)


def gammaincc_(a, x):
    return gamma(a) * gammaincc(a, x)


@pytest.mark.parametrize(
    "s, ndim, k, N, h",
    [
        [-0.5, 1, 0.01, 50, 0.05],
        [-0.5, 1, 1, 50, 0.05],
        [-0.5, 1, 10.0, 50, 0.05],
        [-1, 2, 0.01, 50, 0.05],
        [-1, 2, 1, 50, 0.05],
        [-1, 2, 10.0, 50, 0.05],
        [-1, 4, 0.01, 50, 0.05],
        [-1, 4, 1, 50, 0.05],
        [-1, 4, 10.0, 50, 0.05],
        [-1, 6, 0.01, 50, 0.05],
        [-1, 6, 1, 50, 0.05],
        [-1, 6, 10.0, 50, 0.05],
        [1, 1, 0.01, 150, 10 ** -1.5],
        [1, 1, 1, 150, 10 ** -1.5],
        [1, 1, 10.0, 150, 10 ** -1.5],
        [1, 2, 0.01, 150, 10 ** -1.5],
        [1, 2, 1, 150, 10 ** -1.5],
        [1, 2, 10.0, 150, 10 ** -1.5],
        [1, 3, 0.01, 150, 10 ** -1.5],
        [1, 3, 1, 150, 10 ** -1.5],
        [1, 3, 10.0, 150, 10 ** -1.5],
        [-1, 3, 0.01, 150, 10 ** -1.5],
        [-1, 3, 1, 150, 10 ** -1.5],
        [-1, 3, 10.0, 150, 10 ** -1.5],
    ],
)
def test_powerlaw(s, ndim, k, N, h):
    ht = SymmetricFourierTransform(ndim=ndim, N=N, h=h)
    ans = ht.transform(lambda x: x ** s, k, False, False)

    nu = ndim / 2.0 - 1
    s += nu
    if nu - s <= 0 and (nu - s) % 2 == 0:
        raise Exception("Can't have a negative integer for gamma")

    anl = (
        (2 * np.pi) ** (ndim / 2.0)
        * 2 ** (s + 1)
        * gamma(0.5 * (2 + nu + s))
        / k ** (s + 2)
        / gamma(0.5 * (nu - s))
        / k ** nu
    )

    print("Numerical Result: ", ans, " (required %s)" % anl)
    assert np.isclose(ans, anl, rtol=1e-3)



r"""
This module provides some tests of the integrator to known integrals.

Note, these are not the transformations, just the plain integrals,  :math:`\int_0^\infty f(x) J_\nu(x) dx`

In this module, we use the get_h function, rather than choosing N and h,
to ensure that this function works.
"""

import pytest

import numpy as np
from scipy.special import gamma, gammainc, gammaincc, k0

from hankel import HankelTransform, SymmetricFourierTransform, get_h


def gammainc_(a, x):
    return gamma(a) * gammainc(a, x)


def gammaincc_(a, x):
    return gamma(a) * gammaincc(a, x)


@pytest.mark.parametrize(
    "f, anl",
    [
        (lambda x: 1, 1),  # Ogata 05
        (lambda x: x / (x ** 2 + 1), k0(1)),  # Ogata 05
        (lambda x: x ** 2, -1),  # wikipedia
        (lambda x: x ** 4, 9),  # wikipedia
        (
            lambda x: 1.0 / np.sqrt(x),  # wikipedia
            2 ** (-0.5) * gamma(-1.5 / 2 + 1) / gamma(1.5 / 2),
        ),
        (lambda x: x / np.sqrt(x ** 2 + 1 ** 2), np.exp(-1)),  # wikipedia
        (
            lambda x: x * np.exp(-0.5 * 2 ** 2 * x ** 2),  # wikipedia
            1.0 / 2 ** 2 * np.exp(-0.5 / 2 ** 2),
        ),
    ],
)
def test_nu0(f, anl):
    ans = get_h(f=f, nu=0, hstart=0.5, atol=0, rtol=1e-3, maxiter=20)[1]
    print("Numerical Result: {ans} (required {anl})".format(ans=ans, anl=anl))
    # we ensure that the answer is within 5e-3 of the true answer,
    # because the algorithm does not ensure that the result is within that
    # tolerance of the true answer, only the previous iteration.
    assert np.isclose(ans, anl, atol=0, rtol=5e-3)


@pytest.mark.parametrize(
    "s, nu, N, h",
    [
        [0, 1, 50, 0.05],
        [0, 2, 50, 0.05],
        [0.5, 1, 50, 0.05],
        [-2, 2, 600, 10 ** -2.6],  # This is pretty finnicky
        [-0.783, 1, 50, 0.05],
    ],
)
def test_nu_varying_powerlaw(s, nu, N, h):
    # For this one we test the transform instead
    ans = get_h(f=lambda x: x ** s, nu=nu, K=1, hstart=0.5, atol=1e-3, rtol=1e-3)[1]

    anl = 2 ** (s + 1) * gamma(0.5 * (2 + nu + s)) / gamma(0.5 * (nu - s))
    print("Numerical Result: {ans} (required {anl})".format(ans=ans, anl=anl))
    assert np.isclose(ans, anl, rtol=1e-3, atol=1e-3)


@pytest.mark.parametrize(
    "s, nu, N, h", [[0.5, 1, 50, 0.05], [0.783, 1, 50, 0.05], [1.0, 0.5, 500, 0.01]],
)
def test_nu_varying_gamma_mod(s, nu, N, h):
    ans = get_h(
        f=lambda x: x ** (nu - 2 * s + 1) * gammainc_(s, x ** 2),
        nu=nu,
        hstart=0.5,
        atol=1e-3,
        rtol=1e-3,
    )[1]

    anl = 0.5 ** (2 * s - nu - 1) * gammaincc_(1 - s + nu, 0.25)

    print("Numerical Result: {ans} (required {anl})".format(ans=ans, anl=anl))
    assert np.isclose(ans, anl, rtol=1e-3, atol=1e-3)


def test_too_large_h():
    with pytest.raises(ValueError):
        get_h(f=lambda x: 1, nu=0, hstart=2)


def test_never_converges_high():
    with pytest.raises(Exception):
        get_h(f=lambda x: x ** 6, nu=0, maxiter=15)


def test_never_converges_low():
    with pytest.raises(Exception):
        get_h(f=lambda x: x ** -6, nu=0, maxiter=15)


def test_xrange():
    x1 = HankelTransform.xrange_approx(h=0.5, nu=0)
    x2 = HankelTransform.xrange_approx(h=0.1, nu=0)

    assert x1[0] > x2[0]
    assert x1[1] < x2[1]

    x1 = SymmetricFourierTransform.xrange_approx(h=0.5, ndim=2)
    x2 = SymmetricFourierTransform.xrange_approx(h=0.1, ndim=2)

    assert x1[0] > x2[0]
    assert x1[1] < x2[1]


def test_final_term_amplitude():
    g1 = HankelTransform.final_term_amplitude(f=lambda x: x ** -4, h=0.5)
    g2 = HankelTransform.final_term_amplitude(f=lambda x: x ** -4, h=0.1)
    assert g1 > g2

    g1 = SymmetricFourierTransform.final_term_amplitude(f=lambda x: x ** -4, h=0.5)
    g2 = SymmetricFourierTransform.final_term_amplitude(f=lambda x: x ** -4, h=0.1)
    assert g1 > g2



"""
This module provides some tests of the API of hankel.

That is, it tests stuff like whether one can input scalars/arrays, and whether the outputs of functions have the correct
number of entries etc.
"""

import pytest

import numpy as np

from hankel import HankelTransform


def test_array_n():
    with pytest.raises(ValueError):
        HankelTransform(N=np.linspace(100, 200, 10))


def test_array_h():
    with pytest.raises(TypeError):
        HankelTransform(h=np.logspace(-3, 0, 10))


def test_array_nu():
    with pytest.raises(ValueError):
        HankelTransform(nu=np.linspace(0, 1, 2))


def test_k_array():
    k = np.logspace(-3, 3, 10)
    ht = HankelTransform(N=50)
    res = ht.transform(lambda x: 1.0 / x, k, False)
    assert len(res) == 10


def test_k_scalar():
    k = 1
    ht = HankelTransform(N=50)
    res = ht.transform(lambda x: 1.0 / x, k, False)
    assert np.isscalar(res)


def test_ret_err():
    ht = HankelTransform(N=50)
    res, err = ht.integrate(lambda x: 1.0 / x, True)


def test_ret_cumsum():
    ht = HankelTransform(N=50)
    res, cumsum = ht.integrate(lambda x: 1.0 / x, False, True)


def test_ret_err_and_cumsum():
    ht = HankelTransform(N=50)
    res, err, cumsum = ht.integrate(lambda x: 1.0 / x, True, True)


def test_equivalence_of_integrate_and_transform():
    ht = HankelTransform(N=50)
    intg = ht.integrate(lambda x: 1, False)
    tr = ht.transform(lambda x: 1.0 / x, ret_err=False)
    assert intg == tr



r"""
General quadrature method for Hankel transformations.

Based on the algorithm provided in
H. Ogata, A Numerical Integration Formula Based on the Bessel Functions,
Publications of the Research Institute for Mathematical Sciences,
vol. 41, no. 4, pp. 949-970, 2005.
"""

import warnings
from builtins import super

import numpy as np
from scipy.integrate import quad

from hankel.tools import (
    d_psi,
    dim_to_nu,
    fourier_norm,
    get_x,
    j_lim,
    kernel,
    roots,
    safe_power,
    weight,
)


class HankelTransform:
    r"""
    The basis of the Hankel Transformation algorithm by Ogata 2005.

    This algorithm is used to solve the equation
    :math:`\int_0^\infty f(x) J_\nu(x) dx`
    where :math:`J_\nu(x)` is a Bessel function of the first kind of order
    :math:`\nu`, and :math:`f(x)` is an arbitrary (slowly-decaying) function.

    The algorithm is presented in
    H. Ogata, A Numerical Integration Formula Based on the Bessel Functions,
    Publications of the Research Institute for Mathematical Sciences,
    vol. 41, no. 4, pp. 949-970, 2005.

    This class provides a method for directly performing this integration,
    and also for doing a Hankel Transform.

    Parameters
    ----------
    nu : scalar, optional
        The order of the bessel function (of the first kind) J_nu(x)
    N : int, optional, default = ``pi/h``
        The number of nodes in the calculation. Generally this must increase
        for a smaller value of the step-size h. Default value is based on where
        the series will truncate according
        to the double-exponential convergence to
        the roots of the Bessel function.
    h : float, optional
        The step-size of the integration.
    alt : bool, optional
        Whether to use the alternative definition of the hankel transform.
        should be used. Default: False
    """

    def __init__(self, nu=0, N=None, h=0.05, alt=False):

        N = int(np.pi / h) if N is None else N
        if not np.isscalar(N):
            raise ValueError("N must be a scalar")
        if not np.isscalar(h):
            raise ValueError("h must be a scalar")
        if not np.isscalar(nu):
            raise ValueError("nu must be a scalar")
        if nu < -0.5:
            raise ValueError("nu must be at least -1/2")

        self._nu = nu
        self._h = h
        self._zeros = roots(N, nu)
        self.x = get_x(h, self._zeros)
        self.kernel = kernel(self.x, nu, alt)
        self.w = weight(nu, self._zeros)
        self.dpsi = d_psi(h * self._zeros)
        self.alt = alt

        # Some quantities only useful in the FourierTransform
        self._r_power = 0 if alt else 1
        self._k_power = 0
        # initialize the factors of the series
        self._factor = None

    @property
    def nu(self):
        """Order of the hankel transform."""
        return self._nu

    @property
    def _series_fac(self):
        """Factors for the series."""
        if self._factor is None:
            self._factor = np.pi * self.w * self.kernel * self.dpsi
        return self._factor

    def _k(self, k):
        return np.array(k)

    def _norm(self, inverse=False):
        r"""Scalar normalisation of the transform. Identically 1."""
        return 1

    def _get_series(self, f, k=1):
        with np.errstate(divide="ignore"):  # numpy safely divids by 0
            args = np.divide.outer(self.x, k).T  # x = r*k
        return self._series_fac * f(args) * safe_power(self.x, self._r_power)

    def transform(self, f, k=1, ret_err=True, ret_cumsum=False, inverse=False):
        r"""
        Do the Hankel-transform of the function f.

        Parameters
        ----------
        f : callable
            A function of one variable, representing :math:`f(x)`
        ret_err : boolean, optional, default = True
            Whether to return the estimated error
        ret_cumsum : boolean, optional, default = False
            Whether to return the cumulative sum

        Returns
        -------
        ret : array-like
            The Hankel-transform of f(x) at the provided k. If
            `k` is scalar, then this will be scalar.
        err : array-like
            The estimated error of the approximate integral, at every `k`.
            It is merely the last term in the sum.
            Only returned if `ret_err=True`.
        cumsum : array-like
            The total cumulative sum, for which the last term
            is itself the transform.
            One can use this to check whether the integral is converging.
            Only returned if `ret_cumsum=True`

        Notes
        -----
        The Hankel transform is defined as

        .. math:: F(k) = \int_0^\infty r f(r) J_\nu(kr) dr.

        Or in the alternative case with ``alt=True``:

        .. math:: F(k) = \int_0^\infty f(r) \sqrt{kr} J_\nu(kr) dr.

        The inverse transform is identical (swapping *k* and *r* of course).
        """
        # The following allows for a re-scaling of k when doing FT's.
        k_scalar = np.isscalar(k)
        k = self._k(k)
        # k = zero here
        k_0 = np.isclose(k, 0)
        kn0 = np.invert(k_0)
        k_tmp = k[kn0]
        # The following renormalises by the fourier dual to some power
        knorm = safe_power(k_tmp, self._k_power + self._r_power + 1)
        # The following is the scalar normalisation of the transform
        # The basic transform has a norm of 1.
        # But when doing FT's, this depends on the dimensionality.
        norm = self._norm(inverse)
        # calculate the result for non zero k (int k -> real ret)
        ret = np.empty_like(k, float) if np.isrealobj(k) else np.empty_like(k)
        summation = self._get_series(f, k_tmp)
        ret[kn0] = np.array(norm * np.sum(summation, axis=-1) / knorm)

        # care about k=0
        ret_0 = 0
        err_0 = 0
        if np.any(k_0):
            # limit of J(nu, 0) considering powers of k
            alt_pow = 0.5 if self.alt else 0  # in alt. def sqrt(rk) involved
            nu_th = self._k_power - alt_pow  # threshold

            if np.isclose(self.nu, nu_th):
                lim_r_pow = self._r_power + alt_pow + self.nu
                int_fac = j_lim(self.nu) * norm

                def integrand(r):
                    return f(r) * safe_power(r, lim_r_pow)

                int_res = quad(integrand, 0, np.inf)
                ret_0 = int_res[0] * int_fac
                err_0 = int_res[1] * int_fac
            elif self.nu < nu_th:
                ret_0 = np.nan
            ret[k_0] = ret_0

        if k_scalar:
            ret = ret.item()

        if ret_err:
            err = np.empty_like(ret)
            err[kn0] = norm * np.take(summation, -1, axis=-1) / knorm
            err[k_0] = err_0
        if ret_cumsum:
            cumsum = np.empty(np.shape(self.x) + np.shape(ret), np.array(ret).dtype)
            cumsum[:, kn0] = norm * np.cumsum(summation, axis=-1).T / knorm
            cumsum[:, k_0] = ret_0

        if ret_err and ret_cumsum:
            return ret, err, cumsum
        if ret_err:
            return ret, err
        if ret_cumsum:
            return ret, cumsum
        return ret

    def integrate(self, f, ret_err=True, ret_cumsum=False):
        r"""
        Do the Hankel-type integral of the function f.

        This is *not* the Hankel transform, but rather the simplified
        integral, :math:`\int_0^\infty f(x) J_\nu(x) dx` , equivalent to the
        transform of :math:`f(r)/r` at *k=1*.

        Parameters
        ----------
        f : callable
            A function of one variable, representing :math:`f(x)`
        ret_err : boolean, optional, default = True
            Whether to return the estimated error
        ret_cumsum : boolean, optional, default = False
            Whether to return the cumulative sum

        Returns
        -------
        ret : float
            The Hankel integral of f(x).
        err : float
            The estimated error of the approximate integral. It is merely the last term
            in the sum. Only returned if `ret_err=True`.
        cumsum : array-like
            The total cumulative sum, for which the last term
            is itself the integral. One can use this to check whether the integral is
            converging. Only returned if `ret_cumsum=True`

        See Also
        --------
        transform :
            The Hankel transform (this function calls :func:`transform` with ``k=1`` and
            ``f(x) = f(x)/x``.
        """
        return self.transform(
            f=(lambda x: f(x) / np.sqrt(x)) if self.alt else (lambda x: f(x) / x),
            k=1,
            ret_err=ret_err,
            ret_cumsum=ret_cumsum,
            inverse=False,
        )

    def xrange(self, k=1):
        """
        Tuple giving (min,max) x value evaluated by f(x).

        Parameters
        ----------
        k : array-like, optional
            Scales for the transformation. Leave as 1 for an integral.

        See Also
        --------
        :meth:`xrange_approx` :
            An approximate version of this method which is a classmethod.
        """
        return np.array([self.x.min() / np.max(k), self.x.max() / np.min(k)])

    @classmethod
    def xrange_approx(cls, h, nu, k=1):
        """
        Tuple giving approximate (min,max) x value evaluated by f(x/k).

        Operates under the assumption that N = 3.2/h.

        Parameters
        ----------
        h : float
            The resolution parameter of the Hankel integration
        nu : float
            Order of the integration/transform
        k : array-like, optional
            Scales for the transformation. Leave as 1 for an integral.

        See Also
        --------
        xrange :
            The actual x-range under a given choice of parameters.
        """
        r = roots(1, nu)[0]
        return np.array([np.pi ** 2 * h * r ** 2 / 2 / k, np.pi * np.pi / h / k])

    @classmethod
    def final_term_amplitude(cls, f, h, k=None, *args, **kwargs):
        """
        Get the amplitude of the last term in cumulative sum.

        The absolute value of the non-oscillatory component
        of the summed series' last term, up to a scaling constant.
        This can be used to get the sign of the slope of the amplitude with h.

        Parameters
        ----------
        f : callable
            The function to integrate/transform
        h : float
            The resolution parameter of the hankel integration
        k : float or array-like, optional
            The scale at which to evaluate the transform.
            If None, assume an integral.

        Returns
        -------
        float :
            The value of G, the amplitude of the final term in the series' sum.
        """
        if k is None:
            return np.sqrt(2 * h / np.pi) * f(np.pi * np.pi / h)
        return np.sqrt(np.pi / (2 * h)) * f(np.pi * np.pi / h / k)

    @classmethod
    def G(cls, f, h, k=None, *args, **kwargs):
        """
        Alias of :meth:`final_term_amplitude`.

        .. deprecated:: Deprecated as of v1. Will be removed in v1.2.
        """
        warnings.warn(
            "Using G has been deprecated and will be removed in v1.2. Please use final_term_amplitude instead.",
            category=DeprecationWarning,
        )
        return cls.final_term_amplitude(f, h, k=k, *args, **kwargs)

    @classmethod
    def slope_of_last_term(cls, f, h, *args, **kwargs):
        """Get the slope (up to a constant) of the last term of the series with h.

        Parameters
        ----------
        f : callable
            The function to integrate/transform
        h : float
            The resolution parameter of the hankel integration

        Other Parameters
        ----------------
        args, kwargs :
            All other parameters are passed through to :func:`final_term_amplitude`.

        Returns
        -------
        float :
            The derivative of the last term of the series with h.
        """
        return cls.final_term_amplitude(
            f, h, *args, **kwargs
        ) - cls.final_term_amplitude(f, h / 1.1, *args, **kwargs)

    @classmethod
    def deltaG(cls, f, h, *args, **kwargs):
        """Alias of :meth:`slope_of_last_term`.

        .. deprecated:: Deprecated as of v1. Will be removed in v1.2.
        """
        warnings.warn(
            "Using deltaG has been deprecated and will be removed in v1.2. Please use "
            "slope_of_last_term instead.",
            category=DeprecationWarning,
        )
        return cls.slope_of_last_term(f, h, *args, **kwargs)


class SymmetricFourierTransform(HankelTransform):
    r"""
    Fourier Transform of a radially symmetric function in arbitrary dimensions.

    Parameters
    ----------
    ndim : int
        Number of dimensions the transform is in.
    a, b : float, default 1
        This pair of values defines the Fourier convention used
        (see Notes below for details)
    N : int, optional
        The number of nodes in the calculation. Generally this must increase
        for a smaller value of the step-size h.
    h : float, optional
        The step-size of the integration.
    alt : bool, optional
        State if the alternative definition of the hankel transform
        should be used. Default: False

    Notes
    -----
    We allow for arbitrary Fourier convention, according to the scheme in
    http://mathworld.wolfram.com/FourierTransform.html.
    That is, we define the forward and inverse *n*-dimensional transforms
    respectively as

    .. math:: F(k) = \sqrt{\frac{|b|}{(2\pi)^{1-a}}}^n
              \int f(r) e^{i b\mathbf{k}\cdot\mathbf{r}} d^n\mathbf{r}

    and

    .. math:: f(r) = \sqrt{\frac{|b|}{(2\pi)^{1+a}}}^n
              \int F(k) e^{-i b\mathbf{k}\cdot\mathbf{r}} d^n \mathbf{k}.

    By default, we set both *a* and *b* to 1,
    so that the forward transform has a normalisation of unity.

    In this general sense, the forward and inverse Hankel transforms are
    respectively

    .. math:: F(k) = \sqrt{\frac{|b|}{(2\pi)^{1-a}}}^n
              \frac{(2\pi)^{n/2}}{(bk)^{n/2-1}}
              \int_0^\infty r^{n/2-1} f(r) J_{n/2-1}(bkr) r dr

    and

    .. math:: f(r) = \sqrt{\frac{|b|}{(2\pi)^{1+a}}}^n
              \frac{(2\pi)^{n/2}}{(br)^{n/2-1}}
              \int_0^\infty k^{n/2-1} f(k) J_{n/2-1}(bkr) k dk.

    """

    def __init__(self, ndim=2, a=1, b=1, N=None, h=0.05, alt=True):

        super().__init__(nu=dim_to_nu(ndim), N=N, h=h, alt=alt)

        self.ndim = ndim
        self.fourier_norm_a = a
        self.fourier_norm_b = b
        self._r_power = (ndim - 1) / 2.0 if alt else ndim / 2.0
        self._k_power = (ndim - 1) / 2.0 if alt else ndim / 2.0 - 1

    def _norm(self, inverse=False):
        r"""
        Scalar normalisation of the transform.

        Taking into account Fourier conventions and a possible inversion.
        """
        return (2 * np.pi) ** (self.ndim / 2.0) * fourier_norm(
            self.fourier_norm_a, self.fourier_norm_b, self.ndim, inverse
        )

    def _k(self, k):
        """Substitution for k."""
        return np.array(self.fourier_norm_b * np.array(k))

    @classmethod
    def xrange_approx(cls, h, ndim, k=1):
        """
        Tuple giving approximate (min,max) x value evaluated by f(x/k).

        Operates under the assumption that N = pi/h.

        Parameters
        ----------
        h : float
            The resolution parameter of the Hankel integration
        ndim : float
            Number of dimensions of the transform.
        k : array-like, optional
            Scales for the transformation. Leave as 1 for an integral.

        See Also
        --------
        xrange:  the actual x-range under a given choice of parameters.
        """
        return HankelTransform.xrange_approx(h, dim_to_nu(ndim), k)

    @classmethod
    def final_term_amplitude(cls, f, h, k=None, ndim=2):
        """
        Get the amplitude of the last term in cumulative sum.

        The absolute value of the non-oscillatory component
        of the summed series' last term, up to a scaling constant.
        This can be used to get the sign of the slope of the amplitude with h.

        Parameters
        ----------
        f : callable
            The function to integrate/transform
        h : float
            The resolution parameter of the hankel integration
        k : float or array-like, optional
            The scale at which to evaluate the transform.
            If None, assume an integral.
        ndim : float
            The number of dimensions of the transform

        Returns
        -------
        float :
            The amplitude of the final term in the sum.
        """
        if k is None:
            return HankelTransform.G(f, h, k)
        fmax = f(cls.xrange_approx(h, ndim, k)[-1])
        return (np.pi / h) ** ((ndim - 1) / 2.0) * fmax

    @classmethod
    def G(cls, f, h, k=None, ndim=2):
        """
        Info about the last term in the series.

        .. deprecated:: Deprecated as of v1. Will be removed in v1.2.
        """
        warnings.warn(
            "Using G has been deprecated and will be removed in v1.2. Please use final_term_amplitude instead.",
            category=DeprecationWarning,
        )
        return cls.G(f, h, k, ndim)



"""hankel: Hankel Transformations using method of Ogata 2005."""

import io
import os
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8"),
    ) as fp:
        return fp.read()


CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: Unix",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering",
    "Topic :: Utilities",
]

with open("requirements_dev.txt") as fl:
    test_req = fl.readlines()

with open("docs/requirements.txt") as fl:
    doc_req = fl.readlines()

# use first version of req. to support py35
setup(
    name="hankel",
    install_requires=["numpy>=1.9.3", "scipy>=0.16.1", "mpmath>=1.0.0"],
    classifiers=CLASSIFIERS,
    python_requires=">=3.5",
    packages=["hankel"],
    description="Hankel Transformations using method of Ogata 2005",
    long_description=read("README.rst"),
    author="Steven Murray",
    author_email="steven.murray@curtin.edu.au",
    license="MIT",
    extras_require={"dev": test_req + doc_req, "tests": test_req, "docs": doc_req},
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    url="https://github.com/steven-murray/hankel",
)



"""
General quadrature method for Hankel transformations.

Based on the algorithm provided in
H. Ogata, A Numerical Integration Formula Based on the Bessel Functions,
Publications of the Research Institute for Mathematical Sciences,
vol. 41, no. 4, pp. 949-970, 2005.

.. currentmodule:: hankel

Classes
^^^^^^^

.. autosummary::
   HankelTransform
   SymmetricFourierTransform

Functions
^^^^^^^^^

.. autosummary::
   get_h

"""
from pkg_resources import DistributionNotFound, get_distribution

from hankel.hankel import HankelTransform, SymmetricFourierTransform
from hankel.tools import get_h

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: nocover
    # package is not installed
    pass


__all__ = ["HankelTransform", "SymmetricFourierTransform", "get_h"]
